@@ -44,7 +44,6 @@ const {
   isStackOverflowError,
   codes: {
     ERR_CONSOLE_WRITABLE_STREAM,
-    ERR_INVALID_ARG_VALUE,
     ERR_INCOMPATIBLE_OPTION_PAIR,
   },
 } = require('internal/errors');
@@ -52,6 +51,7 @@ const {
   validateArray,
   validateInteger,
   validateObject,
+  validateOneOf,
 } = require('internal/validators');
 const { previewEntries } = internalBinding('util');
 const { Buffer: { isBuffer } } = require('buffer');
@@ -135,8 +135,7 @@ function Console(options /* or: stdout, stderr, ignoreErrors = true */) {
     throw new ERR_CONSOLE_WRITABLE_STREAM('stderr');
   }
 
-  if (typeof colorMode !== 'boolean' && colorMode !== 'auto')
-    throw new ERR_INVALID_ARG_VALUE('colorMode', colorMode);
+  validateOneOf(colorMode, 'colorMode', ['auto', true, false]);
 
   if (groupIndentation !== undefined) {
     validateInteger(groupIndentation, 'groupIndentation',
